#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <fcntl.h>
#include <glob.h>
#include <pwd.h>
#include <utime.h>
#include <sys/uio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <netinet/in.h>

#include "sftp.h"

#define SCRWIDTH (scrwidth() - 1)

extern u_int32_t version;

int
next_token(char *str, char **arg) {
	char *s, *orig, *startq;
	int escaped = 0, quoted = 0;

	orig = s = str;
	startq = NULL;
	while (1) {
		if (*s == 0) {
			*arg = str;
			return (s - orig);
		}
		if (escaped) {
			memmove(str + 1, str, s - str - 1);
			str++;
			escaped = 0;
		}
		else if (quoted) {
			if (*s == '"') {
				memmove(startq + 2, startq + 1, s - startq - 1);
				memmove(str + 2, str, startq - str);
				str += 2;
				quoted = 0;
			}
		}
		else {
			if (*s == '"') {
				quoted = 1;
				startq = s;
			}
			else if (*s == '\\')
				escaped = 1;
			else if (isspace(*s))
				break;
		}
		s++;
	}
	*s++ = 0;
	*arg = str;
	return (s - orig);
}

int
tokenize(char *str, char **args, int maxargs) {
	int i;

	for (i = 0; str && *str && i < maxargs - 1; i++) {
		skipspaces(str);
		if (*str == 0)
			break;
		str += next_token(str, &args[i]);
	}
	args[i] = NULL;
	return i;
}

char *
findbasename(char *str) {
	char *base;
	base = strrchr(str, '/');
	if (base == NULL)
		base = str;
	else
		base++;
	return base;
}

int
find_match(char *str, map *action) {
	int i, guess = -1, badguess = 0;
	for (i=0; action[i].str; i++) {
		if (strcmp(action[i].str, str) == 0)
			return i;
		if (strncmp(action[i].str, str, strlen(str)) == 0) {
			if (guess < 0 && !badguess)
				guess = i;
			else {
				badguess = 1;
				guess = -1;
			}
		}
	}
	return guess;
}

int
do_action(char *str, map *action) {
	int index, nargs;
	char *args[MAXARGS];

	if (str[strlen(str)-1] == '\n')
		str[strlen(str)-1] = 0;
	skipspaces(str);
	if (*str == 0)
		return 0;
	if (*str == '!') {
		system(str+1);
		return 0;
        }
	nargs = tokenize(str, args, MAXARGS);
	if (nargs == 0)
		return 0;
	index = find_match(args[0], action);
	if (index >= 0)
		return action[index].func(args, nargs);
	else
		printf ("%s: invalid command\n", args[0]);
	return -1;
}

int
readn(int fd, void *buf, size_t count) {
	int n, total = 0;

	while (total < count) {
		n = read(fd, (char*)buf+total, count-total);
		if (n == 0)
			return total;
		else
			total += n;
	}
	return total;
}


int
send_message(int sock, message m) {
	struct iovec v[4];
	u_int32_t nlen = htonl(m.len);
	int n;

	v[0].iov_base = (void *)&m.channel;
	v[0].iov_len = 1;

	v[1].iov_base = (void *)&m.command;
	v[1].iov_len = 1;

	v[2].iov_base = (void *) &nlen;
	v[2].iov_len = 4;

	v[3].iov_base = (void *)m.data;
	v[3].iov_len = m.len;

	n = writev(sock, v, 4);
	if (n < 0)
		return -1;
	return 0;

}

int
recv_message(int sock, message *m) {
	int ret;
	unsigned char buf[6];

	ret = readn(sock, buf, 6);
	if (ret < 6)
		return (-1);
	m->channel = buf[0];
	m->command = buf[1];
	memcpy(&m->len, &buf[2], 4);
	m->len = ntohl(m->len);
	
	if (m->len > 0) {
		m->data = malloc(m->len);
		ret = readn(sock, m->data, m->len);
		if (ret < m->len) {
			free(m->data);
			memset(m, 0, sizeof(message));
			return (-1);
		}
	}
	else
		m->data = NULL;
	return 0;
}

int
query_message(int sock, message *m) {
	int ret;
	struct timeval tv;
	fd_set fds;

	memset(&tv, 0, sizeof(tv));
	FD_ZERO(&fds);
	FD_SET(sock, &fds);
	ret = select (sock+1, &fds, NULL, NULL, &tv);
	if (ret == 0) {
		m->data = NULL;
		return -1;
	}
	else
		return recv_message(sock, m);
}

message
_message(u_int8_t command, void *data, u_int32_t len) {
	return _data_message(0, command, data, len);
}

message
_data_message(u_int8_t channel, u_int8_t command, void *data, u_int32_t len) {
	message m;
	m.channel = channel;
	m.command = command;
	m.len = len;
	m.data = data;
	return m;
}

static int
do_send1file(int sock, char *file) {
	struct stat statbuf;
	u_int32_t size, mtime;
	u_int16_t mode;
	char modestr[9];
	int n, fd;
	char buf[BUFSIZE];
	char *base;
	message m;
	char *str;
	u_int8_t c = 1;
	u_int32_t skipped = 0;

	if (stat(file, &statbuf) || !S_ISREG(statbuf.st_mode)) {
		str = "File doesn't exist or is not a regular file";
		send_message(sock, _data_message(c, ERROR, str, strlen(str)+1));
		return -1;
	}
	fd = open(file, O_RDONLY);
	if (fd < 0) {
		str = "File could not be opened";
		send_message(sock, _data_message(c, ERROR, str, strlen(str)+1));
		return -1;
	}

	base = findbasename(file);
	mode = statbuf.st_mode & (S_IRWXU|S_IRWXG|S_IRWXO);
	modetostr(mode, modestr);
	size = htonl(statbuf.st_size);
	mtime = htonl(statbuf.st_mtime);

	send_message(sock, _message(SENDFILE, &c, 1));
	send_message(sock, _data_message(c, FILENAME, base, strlen(base)+1));
	recv_message(sock, &m);

	if (!(m.command == FILEOK && m.len == 1 && *(char *)m.data != 0)) {
		xfree(m.data);
		return -1;
	}

	xfree(m.data);
	send_message(sock, _data_message(c, FILESIZE, &size, 4));
	send_message(sock, _data_message(c, FILEMODE, modestr, 9));
	if (version >= SFTP_VERSION(0,7,0))
		send_message(sock, _data_message(c, FILETIME, &mtime, 4));

	if (skipped > 0) {
		lseek(fd, skipped, SEEK_SET);
		skipped = htonl(skipped);
		send_message(sock, _data_message(c, SKIPBYTES, &skipped, 4));
	}
	while ((n = read(fd, buf, BUFSIZE)) > 0) {
		send_message(sock, _data_message(c, DATA, buf, n));
	}
	send_message(sock, _data_message(c, ENDDATA, NULL, 0));
	close (fd);
	return 0;
}

int
do_sendfile(int sock, char *file) {
	int ret, i;
	glob_t globbuf;

	if (*file == 0) {
		send_message(sock, _message(NOFILEMATCH, NULL, 0));
		return 0;
	}
	ret = glob(file, 0, NULL, &globbuf);
	if (ret != 0)
		return -1;
	for (i = 0; i < globbuf.gl_pathc; i++)
		do_send1file (sock, globbuf.gl_pathv[i]);
	globfree(&globbuf);
	return i;
}

int
do_recvfile(int sock, u_int8_t channel, int resume) {
	u_int32_t size = 0, mtime = 0;
	u_int16_t mode;
	char *file = NULL;
	int fd = -1, total = 0;
	u_int8_t status;
	message m;
	int namelen = 0, dots = 0, newdots, i;
	u_int32_t startsize = 0;

	while (1) {
		recv_message(sock, &m);
		switch (m.command) {
			case SENDFILE:
				channel = *(u_int8_t*)m.data;
			case FILENAME:
				file = strdup(m.data);
				fd = open(file, O_WRONLY|O_CREAT, 0644);
				fputs(file, stdout);
				fputs(": ", stdout);
				fflush(stdout);
				namelen = (strlen(file) + 2) % SCRWIDTH;
				status = (fd >= 0);
				if (status > 0 && resume) {
					struct stat statbuf;
					u_int32_t nsize;
					fstat(fd, &statbuf);
					startsize = statbuf.st_size;
					lseek(fd, startsize, SEEK_SET);
					nsize = htonl(startsize);
					send_message(sock,
						     _data_message(channel,
								   SKIPBYTES,
								   &nsize, 4));
				}
				else {
					send_message(sock,
						     _data_message(channel,
								   FILEOK,
								   &status, 1));
				}
				if (status == 0) {
					free(file);
					return -1;
				}
				break;
			case FILESIZE:
				if (fd < 0 || m.len != 4)
					break;
				size = ntohl(*(u_int32_t *)m.data);
				break;
			case FILEMODE:
				if (fd < 0 || m.len != 9)
					break;
				mode = strtomode((char *)m.data);
				fchmod(fd, mode);
				break;
			case FILETIME:
				if (fd < 0 || m.len != 4)
					break;
				mtime = ntohl(*(u_int32_t *)m.data);
				break;
			case DATA:
				if (fd < 0)
					break;
				write(fd, m.data, m.len);
				total += m.len;
				if ((size - startsize) == 0)
					break;
				newdots = ((float)total/(size - startsize)) *
					   (SCRWIDTH - namelen);
				for (i = 0; i < newdots - dots; i++)
					putchar('.');
				fflush(stdout);
				dots = newdots;
				break;
			case ENDDATA:
				close(fd);
				if (mtime != 0) {
					struct utimbuf ubuf;
					ubuf.actime = time(NULL);
					ubuf.modtime = mtime;
					utime(file, &ubuf);
				}
				putchar('\n');
				fflush(stdout);
				free(file);
				return size - startsize;
		}
		xfree(m.data);
	}
}

struct timeval
timediff(struct timeval tv1, struct timeval tv2) {
	struct timeval tv;
	tv.tv_sec = tv1.tv_sec - tv2.tv_sec;
	tv.tv_usec = tv1.tv_usec - tv2.tv_usec;
	if (tv.tv_usec < 0) {
		tv.tv_usec+=1000000;
		tv.tv_sec--;
	}
	return tv;
}

sftp_channel *
new_channel() {
	sftp_channel *channel = (sftp_channel *) malloc(sizeof(sftp_channel));
	memset (channel, 0, sizeof(sftp_channel));
	return channel;
}

u_int16_t
strtomode(char *str) {
	u_int16_t mode = 0;
	if (str[0] == 'r') mode |= S_IRUSR;
	if (str[1] == 'w') mode |= S_IWUSR;
	if (str[2] == 'x') mode |= S_IXUSR;
	if (str[3] == 'r') mode |= S_IRGRP;
	if (str[4] == 'w') mode |= S_IWGRP;
	if (str[5] == 'x') mode |= S_IXGRP;
	if (str[6] == 'r') mode |= S_IROTH;
	if (str[7] == 'w') mode |= S_IWOTH;
	if (str[8] == 'x') mode |= S_IXOTH;
	return mode;
}

void
modetostr(u_int16_t mode, char *str) {
	strncpy(str, "rwxrwxrwx", 9);
	if (!(mode & S_IRUSR)) str[0] = '-';
	if (!(mode & S_IWUSR)) str[1] = '-';
	if (!(mode & S_IXUSR)) str[2] = '-';
	if (!(mode & S_IRGRP)) str[3] = '-';
	if (!(mode & S_IWGRP)) str[4] = '-';
	if (!(mode & S_IXGRP)) str[5] = '-';
	if (!(mode & S_IROTH)) str[6] = '-';
	if (!(mode & S_IWOTH)) str[7] = '-';
	if (!(mode & S_IXOTH)) str[8] = '-';
}

int
scrwidth() {
#ifdef TIOCGWINSZ
	static int width = -1;
	int ret;
	struct winsize win;

	if (width != -1)
		return width;
	ret = ioctl(1, TIOCGWINSZ, (char *)&win);
	if (ret != 0)
		width = 80;
	else
		width = win.ws_col;
	return width;
#else
	return 80;
#endif
}

/* Note - this is _not_ thread safe. */
char *
tildeexpand(char *path) {
	static char newpath[256];
	struct passwd *pwd;
	char *str;

	if (path[0] != '~')
		return path;
	str = strchr(path, '/');
	if (str != NULL)
		*str++ = 0;
	if (strcmp(path, "~") == 0)
		pwd = getpwuid(getuid());
	else
		pwd = getpwnam(path + 1);
	if (pwd == NULL)
		return NULL;
	if (str == NULL)
		return pwd->pw_dir;
	if (strlen(pwd->pw_dir) + strlen(str) + 2 > sizeof(newpath))
		return NULL;
	sprintf(newpath, "%s/%s", pwd->pw_dir, str);
	return newpath;
}
